#include <assert.h>
#include <errno.h>
#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "options.h"
#include "util.h"

/* version string includes compile time and date for convenience */
#define MAKE_VERSION(ver) "Hermit host-side v" #ver " compiled "__TIME__", "__DATE__" (CVS: $Id: util.c,v 1.24 2001/03/30 05:34:31 miket Exp $)";

static const char hermit_version[] = MAKE_VERSION("1.3-armadillo-2");

/* name of this executable, for errors and warnings */
const char *execname;

void set_execname(const char *execpath)
{
	const char *p = strrchr(execpath, '/');
	execname = p ? ++p : execpath;
}

const char *get_version(void)
{
	return hermit_version;
}

void perror_xe(const char *errstr)
{
	fprintf(stderr, "%s: ", execname);
	perror(errstr);
	exit(EXIT_FAILURE);
}

void perror_xu(const char *errstr)
{
	fprintf(stderr, "%s: ", execname);
	perror(errstr);
	usage_and_exit();
}

void per_line(const char *pathname, iter_func_t *func, void *arg)
{
	char buf [LINE_LENGTH];
	FILE *f;

	assert(pathname);
	assert(func);
	if (!(f = fopen(pathname, "r")))
		perror_xe(pathname);
	while (fgets(buf, sizeof buf, f)) {
		if (func(buf, arg) < 0)
			break;
	}
	if (ferror(f))
		perror_xe(pathname);
	fclose(f);
}

FILE *msgout, *panicout, *warnout;

void init_util_output(void)
{
	msgout = stderr;
	panicout = stderr;
	warnout = stderr;
}

int msg(const char *message)
{
	int n;
	assert(msgout);
	if (!opt_verbose)
		return 0;
	n = fprintf(msgout, "%s", message);
	fflush(msgout);
	return n;
}

int msgf(const char *format, ...)
{
	va_list ap;
	int n;

	assert(msgout);
	if (!opt_verbose)
		return 0;
	va_start(ap, format);
	n = vfprintf(msgout, format, ap);
	va_end(ap);
	fflush(msgout);
	return n;
}

void panic(const char *message)
{
	assert(panicout);
	fprintf(panicout, "%s: panic: %s\n", execname, message);
	exit(EXIT_FAILURE);	
}

void panicf(const char *format, ...)
{
	va_list ap;

	assert(panicout);
	fprintf(panicout, "%s: panic: ", execname);
	va_start(ap, format);
	vfprintf(panicout, format, ap);
	va_end(ap);
	fputc('\n', panicout);
	exit(EXIT_FAILURE);	
}

int warn(const char *message)
{
	assert(warnout);
	return fprintf(warnout, "%s: warning: %s", execname, message);
}

int warnf(const char *format, ...)
{
	va_list ap;
	int n;

	assert(warnout);
	n = fprintf(warnout, "%s: warning: ", execname);
	va_start(ap, format);
	n += vfprintf(warnout, format, ap);
	va_end(ap);
	
	return n;
}

/* close or die */
void xclose(int fd)
{
	if (close(fd))
		perror_xe("close");
}

/* read or die */
ssize_t xread(int fd, void *buf, size_t count)
{
	ssize_t retval;
	do {
		retval = read(fd, buf, count);
	} while ((retval < 0) && (errno == EINTR));
	if (retval < 0)
		perror_xe("read");
	return retval;
}

/* read entire buffer or die */
ssize_t xaread(int fd, void *buf, size_t count)
{
	int nread = 1, remain;
	for (remain = count; nread && remain; remain -= nread)
		buf += (nread = xread(fd, buf, remain));
	return count;
}

/* write or die */
ssize_t xwrite(int fd, const void *buf, size_t count)
{
	ssize_t retval;
	do {
		retval = write(fd, buf, count);
	} while ((retval < 0) && (errno == EINTR));
	if (retval < 0)
		perror_xe("write");
	return retval;
}

/* write entire buffer or die */
ssize_t xawrite(int fd, const void *buf, size_t count)
{
	int nwritten, remain;
	for (remain = count; remain; remain -= nwritten)
		buf += (nwritten = xwrite(fd, buf, remain));
	return count;
}

/* malloc or die */
void *xmalloc(size_t size)
{
	void *value = malloc(size);
	if (!value) {
		fprintf(stderr, "%s: virtual memory exhausted\n", execname);
		exit(7);
	}
	return value;
}

/* zero-malloc or die */
void *zmalloc(size_t size)
{
	void *value = calloc(size, 1);
	if (!value) {
		fprintf(stderr, "%s: virtual memory exhausted\n", execname);
		exit(7);
	}
	return value;
}

char *xstrdup(const char *s)
{
	char *value = strdup(s);
	if (!value) {
		fprintf(stderr, "%s: virtual memory exhausted\n", execname);
		exit(7);
	}
	return value;
}
